#include "stdafx.h"

#include "gpuPeteOpenGL2Tweak.h"
#include "GPUPlugin.h"

#include "SafeWrite.h"

#include "gte_accuracy.h"

#include "GPUPatches.h"
#include "xBRZ\hq.h"

GPUPatches::GTEData GPUPatches::m_gtedata;

GPUPatches::tOffset GPUPatches::ooffset3 = nullptr;
GPUPatches::tOffset GPUPatches::ooffset4 = nullptr;

WNDPROC GPUPatches::oldWndProc = nullptr;

u8* GPUPatches::m_RGBABuf;
u8 GPUPatches::m_scale;

GPUPatches::GPUPatches()
{
	m_gtedata.lx0 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A08);
	m_gtedata.lx1 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A0A);
	m_gtedata.lx2 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A0C);
	m_gtedata.lx3 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A0E);

	m_gtedata.ly0 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A10);
	m_gtedata.ly1 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A12);
	m_gtedata.ly2 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A14);
	m_gtedata.ly3 = (s16*)GPUPlugin::Get().GetPluginMem(0x00051A16);

	m_gtedata.vertex[0] = (OGLVertex*)GPUPlugin::Get().GetPluginMem(0x00052220);
	m_gtedata.vertex[1] = (OGLVertex*)GPUPlugin::Get().GetPluginMem(0x00052238);
	m_gtedata.vertex[2] = (OGLVertex*)GPUPlugin::Get().GetPluginMem(0x00052250);
	m_gtedata.vertex[3] = (OGLVertex*)GPUPlugin::Get().GetPluginMem(0x00052268);

	m_gtedata.PSXDisplay_CumulOffset_x = (s16*)GPUPlugin::Get().GetPluginMem(0x00051FFC);
	m_gtedata.PSXDisplay_CumulOffset_y = (s16*)GPUPlugin::Get().GetPluginMem(0x00051FFE);
}

GPUPatches::~GPUPatches()
{
	delete[] m_RGBABuf;
}

int GPUPatches::GetVideoMemoryAMD()
{
	if (!WGLEW_AMD_gpu_association) return 0;

	size_t total_mem_mb = 0;
	if (wglGetGPUIDsAMD && wglGetGPUInfoAMD)
	{
		UINT n = wglGetGPUIDsAMD(0, 0);
		UINT *ids = (UINT*)malloc(n*sizeof(UINT));
		wglGetGPUIDsAMD(n, ids);
		if (ids)
			wglGetGPUInfoAMD(ids[0], WGL_GPU_RAM_AMD, GL_UNSIGNED_INT, sizeof(size_t), &total_mem_mb);
		free(ids);
	}

	PLUGINLOG("AMD GPU VRAM: %uMB", total_mem_mb);
	return (int)total_mem_mb;
};

int GPUPatches::GetVideoMemoryNV()
{
	if (!GLEW_NVX_gpu_memory_info) return 0;

	GLint total_mem_kb = 0;
	glGetIntegerv(GL_GPU_MEMORY_INFO_TOTAL_AVAILABLE_MEMORY_NVX, &total_mem_kb);

	int iVRamSize = total_mem_kb / 1024;

	PLUGINLOG("NV GPU VRAM: %uMB", iVRamSize);
	return iVRamSize;
}

typedef u32(__cdecl* tCheckTextureMemory)();

void GPUPatches::FixMemoryDetection()
{
	// GetVideoMemoryAMD will return 0 on non AMD platforms
	int iVRamSize = GetVideoMemoryAMD();
	if (!iVRamSize) iVRamSize = GetVideoMemoryNV();

	//if (iVRamSize > 1024) iVRamSize = 1024; //clamp memory to 1GB, more does nothing
	if (iVRamSize) SafeWrite<u32>((u32*)GPUPlugin::Get().GetPluginMem(0x00050224), iVRamSize);

	//now recheck usable textures
	tCheckTextureMemory CheckTextureMemory = (tCheckTextureMemory)GPUPlugin::Get().GetPluginMem(0x000448B0);
	CheckTextureMemory();
}

void GPUPatches::EnableVsync(u32 interval)
{
	if (!WGLEW_EXT_swap_control) return;

	if (wglSwapIntervalEXT)
	{
		if (wglSwapIntervalEXT(interval))
			PLUGINLOG("SwapInterval set to %d", interval);
		else
			PLUGINLOG("Fail to set SwapInterval to %d", interval);
	}
}

/*

reversed:

dwActFixes <= dword_100500F4

offset3 <= signed int __cdecl sub_100041B0()
offset4 <= signed int __cdecl sub_100043F0()

word_10051A08 => lx0
word_10051A0A => lx1
word_10051A0C => lx2
word_10051A0E => lx3

word_10051A10 => ly0
word_10051A12 => ly1
word_10051A14 => ly2
word_10051A16 => ly3

flt_10052220 => vertex[0].x
flt_10052238 => vertex[1].x
flt_10052250 => vertex[2].x
flt_10052268 => vertex[3].x

flt_10052224 => vertex[0].y
flt_1005223C => vertex[1].y
flt_10052254 => vertex[2].y
flt_1005226C => vertex[3].y

dword_10051FFC => PSXDisplay_CumulOffset_x_y

*/

void GPUPatches::fix_offsets(u32 _case)
{
	//PLUGINLOG("GTE Accuracy offset%d", _case);
	switch (_case)
	{
	case 4:
		if (getGteVertex(*m_gtedata.lx3, *m_gtedata.ly3, &m_gtedata.vertex[3]->x, &m_gtedata.vertex[3]->y))
		{
			m_gtedata.vertex[3]->x += *m_gtedata.PSXDisplay_CumulOffset_x;
			m_gtedata.vertex[3]->y += *m_gtedata.PSXDisplay_CumulOffset_y;
		}
	case 3:
		if (getGteVertex(*m_gtedata.lx2, *m_gtedata.ly2, &m_gtedata.vertex[2]->x, &m_gtedata.vertex[2]->y))
		{
			m_gtedata.vertex[2]->x += *m_gtedata.PSXDisplay_CumulOffset_x;
			m_gtedata.vertex[2]->y += *m_gtedata.PSXDisplay_CumulOffset_y;
		}
	case 2:
		if (getGteVertex(*m_gtedata.lx1, *m_gtedata.ly1, &m_gtedata.vertex[1]->x, &m_gtedata.vertex[1]->y))
		{
			m_gtedata.vertex[1]->x += *m_gtedata.PSXDisplay_CumulOffset_x;
			m_gtedata.vertex[1]->y += *m_gtedata.PSXDisplay_CumulOffset_y;
		}
	case 1:
		if (getGteVertex(*m_gtedata.lx0, *m_gtedata.ly0, &m_gtedata.vertex[0]->x, &m_gtedata.vertex[0]->y))
		{
			m_gtedata.vertex[0]->x += *m_gtedata.PSXDisplay_CumulOffset_x;
			m_gtedata.vertex[0]->y += *m_gtedata.PSXDisplay_CumulOffset_y;
		}
	}
}

BOOL __cdecl GPUPatches::offset3(void)
{
	BOOL ret = ooffset3();
	fix_offsets(3);
	return ret;
}

BOOL __cdecl GPUPatches::offset4(void)
{
	BOOL ret = ooffset4();
	fix_offsets(4);
	return ret;
}

void GPUPatches::GTEAccuracy()
{
	if (!ooffset3 && !ooffset4)
	{
		//offsetline = signed int __cdecl sub_10003DC0()
		//offset3 = signed int __cdecl sub_100041B0()
		//offset4 = signed int __cdecl sub_100043F0()
		//offsetST = signed int __cdecl sub_10004780()
		//??????? = int __cdecl sub_100048B0()
		//??????? = int __cdecl sub_10004990()

		tOffset _offset3 = (tOffset)GPUPlugin::Get().GetPluginMem(0x000041B0);
		tOffset _offset4 = (tOffset)GPUPlugin::Get().GetPluginMem(0x000043F0);

		MH_CreateHook(_offset3, offset3, reinterpret_cast<void**>(&ooffset3));
		MH_EnableHook(_offset3);

		MH_CreateHook(_offset4, offset4, reinterpret_cast<void**>(&ooffset4));
		MH_EnableHook(_offset4);
	}
}

void GPUPatches::ResHack(u32 _x, u32 _y)
{
	u8 x = (u8)_x;
	u8 y = (u8)_y;

	if (x == 0) x = 1;
	if (y == 0) y = 1;

	PLUGINLOG("ResHack %u %u", x, y);

	u16* nopX = (u16*)GPUPlugin::Get().GetPluginMem(0x0000942E);
	u8* mulX = (u8*)GPUPlugin::Get().GetPluginMem(0x00009431);

	u16* nopY = (u16*)GPUPlugin::Get().GetPluginMem(0x00009449);
	u8* mulY = (u8*)GPUPlugin::Get().GetPluginMem(0x0000944C);

	SafeWrite<u16>(nopX, 0x9090);
	SafeWrite<u16>(nopY, 0x9090);

	SafeWrite<u8>(mulX, x);
	SafeWrite<u8>(mulY, y);
}

LRESULT CALLBACK GPUPatches::TweakWindowProc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
	switch (uMsg)
	{
	case WM_KEYDOWN:
		if (wParam == VK_F1 ||
			wParam == VK_F2 ||
			wParam == VK_F3)
		{
			return DefWindowProc(hWnd, uMsg, wParam, lParam);
		}
	}

	return CallWindowProc(oldWndProc, hWnd, uMsg, wParam, lParam);
}

void GPUPatches::HardcoreMode(HWND hWnd)
{
	if (!oldWndProc)
		oldWndProc = (WNDPROC)SetWindowLong(hWnd, GWL_WNDPROC, (LONG)TweakWindowProc);
}

void (APIENTRY* GPUPatches::oglCopyTexSubImage2D)(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height) = nullptr;
void APIENTRY GPUPatches::Hook_glCopyTexSubImage2D(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLint x, GLint y, GLsizei width, GLsizei height)
{
	PLUGINLOG("glCopyTexSubImage2D 0x%08X %d %d %d %d %d %d %d", target, level, xoffset, yoffset, x, y, width, height);

	if (target != GL_TEXTURE_2D)
		return oglCopyTexSubImage2D(target, level, xoffset, yoffset, x, y, width, height);

	if (width > MAX_TEXTURE_X || height > MAX_TEXTURE_Y)
	{
		PLUGINLOG("glCopyTexSubImage2D !!!Texture too big !!!");
		PLUGINLOG("glCopyTexSubImage2D 0x%08X %d %d %d %d %d %d %d", target, level, xoffset, yoffset, x, y, width, height);
		return oglCopyTexSubImage2D(target, level, xoffset, yoffset, x, y, width, height);
	}

	return oglCopyTexSubImage2D(target, level, xoffset * m_scale, yoffset * m_scale, x, y, width * m_scale, height * m_scale);
}

void (APIENTRY* GPUPatches::oglTexSubImage2D)(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels) = nullptr;
void APIENTRY GPUPatches::Hook_glTexSubImage2D(GLenum target, GLint level, GLint xoffset, GLint yoffset, GLsizei width, GLsizei height, GLenum format, GLenum type, const GLvoid *pixels)
{
	if (target != GL_TEXTURE_2D || format != GL_RGBA || width > MAX_TEXTURE_X || height > MAX_TEXTURE_Y)
	{
		if (width > MAX_TEXTURE_X || height > MAX_TEXTURE_Y)
			PLUGINLOG("glTexSubImage2D !!!Texture too big !!!");
		PLUGINLOG("glTexSubImage2D 0x%08X %d %d %d %d %d 0x%08X 0x%08X %p", target, level, xoffset, yoffset, width, height, format, type, pixels);
		return oglTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, pixels);
	}

	if (format == GL_RGBA)
	{
		scaleXbrz(m_scale, pixels, m_RGBABuf, 32, width, height);
		return oglTexSubImage2D(target, level, xoffset * m_scale, yoffset * m_scale, width * m_scale, height * m_scale, format, type, m_RGBABuf);
	}
	return oglTexSubImage2D(target, level, xoffset, yoffset, width, height, format, type, pixels);
}

void (APIENTRY* GPUPatches::oglTexImage2D)(GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const void *pixels) = nullptr;
void APIENTRY GPUPatches::Hook_glTexImage2D(GLenum target, GLint level, GLint internalformat, GLsizei width, GLsizei height, GLint border, GLenum format, GLenum type, const void *pixels)
{
	// render target 
	if (width == 4096 || height == 4096)
		return oglTexImage2D(target, level, internalformat, width, height, border, format, type, pixels);

	if (target != GL_TEXTURE_2D || format != GL_RGBA || width > MAX_TEXTURE_X || height > MAX_TEXTURE_Y)
	{
		if (width > MAX_TEXTURE_X || height > MAX_TEXTURE_Y)
			PLUGINLOG("glTexImage2D !!!Texture too big !!!");
		PLUGINLOG("glTexImage2D 0x%08X %d %d %d %d %d 0x%08X 0x%08X %p", target, level, internalformat, width, height, border, format, type, pixels);
		return oglTexImage2D(target, level, internalformat, width, height, border, format, type, pixels);
	}

	if (format == GL_RGBA)
	{
		scaleXbrz(m_scale, pixels, m_RGBABuf, 32, width, height);
		return oglTexImage2D(target, level, internalformat, width * m_scale, height * m_scale, border, format, type, m_RGBABuf);
	}
	return oglTexImage2D(target, level, internalformat, width, height, border, format, type, pixels);
}

void GPUPatches::TextureScale(u8 scale)
{
	std::once_flag flag;
	std::call_once(flag, [&]()
	{
		m_scale = scale;

		m_RGBABuf = new u8[MAX_TEXTURE_X * m_scale * MAX_TEXTURE_Y * m_scale * 4];

		MH_CreateHook(glTexImage2D, Hook_glTexImage2D, reinterpret_cast<void**>(&oglTexImage2D));
		MH_EnableHook(glTexImage2D);

		MH_CreateHook(glTexSubImage2D, Hook_glTexSubImage2D, reinterpret_cast<void**>(&oglTexSubImage2D));
		MH_EnableHook(glTexSubImage2D);

		MH_CreateHook(glCopyTexSubImage2D, Hook_glCopyTexSubImage2D, reinterpret_cast<void**>(&oglCopyTexSubImage2D));
		MH_EnableHook(glCopyTexSubImage2D);
	}
	);
}
