
#include "stdafx.h"

#include "PADPlugin.h"
#include "gpuPeteOpenGL2Tweak.h"
#include "GPUPlugin.h"

Context Context::instance;

Context::Context()
{
	MH_Initialize();
}

Context::~Context()
{
	MH_Uninitialize();
}

void Context::LoadPad()
{
	if (!m_config.GetePSXePadPlugin())
		return;
	if (!m_padplugin.Hook())
		return;

	if (!m_config.GetPort1().empty())
		m_padplugin.Load(m_config.GetPort1(), 0);

	if (!m_config.GetPort2().empty())
		m_padplugin.Load(m_config.GetPort2(), 1);

	m_padplugin.OnPADinit();
}

s32 Context::OnGPUinit()
{
	PLUGINLOG("Pete OpenGL2 Tweak Enabled");
	if (m_config.GetGTEAccuracy())
		m_gpupatches.GTEAccuracy();

	if (m_config.GetxBRZScale() > 1)
	{
		u8 scale = m_config.GetxBRZScale();
		if (scale > 5) scale = 5;
		m_gpupatches.TextureScale(scale);
	}

	return GPUPlugin::Get().GPUinit();
}

s32  Context::OnGPUshutdown()
{
	m_padplugin.OnPADshutdown();
	return GPUPlugin::Get().GPUshutdown();
}

s32  Context::OnGPUOpen(HWND hwndGPU)
{
	s32 ret = GPUPlugin::Get().GPUopen(hwndGPU);

	std::once_flag glewInitFlag;
	std::call_once(glewInitFlag, glewInit);

	if (m_config.GetHardcoreMode())
		m_gpupatches.HardcoreMode(hwndGPU);

	if (m_config.GetUseWindow())
	{
		UINT flags = SWP_NOSIZE;
		if (!m_config.GetWindowOnTop()) flags |= SWP_NOZORDER;

		SetWindowPos(hwndGPU, HWND_TOPMOST, m_config.GetWindowX(), m_config.GetWindowY(), 0, 0, flags);
	}

	if (m_config.GetFixMemoryDetection())
		m_gpupatches.FixMemoryDetection();

	if (m_config.GetVSyncInterval())
		m_gpupatches.EnableVsync(m_config.GetVSyncInterval());

	if (m_config.GetGTEAccuracy())
		resetGteVertices();

	if (m_config.GetResHack())
		m_gpupatches.ResHack(m_config.GetMulX(), m_config.GetMulY());

	if (m_config.GetHideCursor())
		while (ShowCursor(FALSE) > -1);

	LoadPad();
	m_padplugin.OnPADopen(hwndGPU);
	return ret;
}

s32  Context::OnGPUclose()
{
	m_padplugin.OnPADclose();

	if (m_config.GetHideCursor()) while (ShowCursor(TRUE) < 0);
	return GPUPlugin::Get().GPUclose();
}

s32 Context::OnGPUtest()
{
	if (m_config.GetePSXePadPlugin())
	{
		if (!m_config.GetPort1().empty())
			m_padplugin.Load(m_config.GetPort1(), 0);

		if (!m_config.GetPort2().empty())
			m_padplugin.Load(m_config.GetPort2(), 1);

		m_padplugin.OnPADinit();

		m_padplugin.OnPADconfigure();
	}
	return GPUPlugin::Get().GPUtest();
}

void  Context::OnGPUaddVertex(s16 sx, s16 sy, s64 fx, s64 fy, s64 fz)
{
	static std::once_flag flag;
	std::call_once(flag, [&]()
	{
		PLUGINLOG("GTE Accuracy Hack Enabled");
	});

	//Uncomment for debug
	//PLUGINLOG("GPUaddVertex: sx = %hu sy = %hu fx = %llu fy = %llu fz = %llu", sx, sy, fx, fy, fz);
	if (m_config.GetGTEAccuracy())
		_GPUaddVertex(sx, sy, fx, fy, fz);
}