
#include "gte_accuracy.h"
#include "GPUPatches.h"
#include "PADPlugin.h"

#define PLUGINLOG(format, ...) printf("TWEAK: "format"\n", __VA_ARGS__)

static const char *ini_filename = ".\\inis\\gpuPeteOpenGL2Tweak.ini";
static const char *libraryName = "PeteOpenGL2 Tweaks";

static const char *dllfilename = "plugins\\gpuPeteOpenGL2.dll";
static const char *pecfilename = "plugins\\gpupec.dll";

#define VERSION_MAJOR 2
#define VERSION_MINOR 0

#define CONFIG_VERSION 6

class Config : NonCopyable
{
public:
    Config()
    {
        IniFile ini;
        ini.Load(ini_filename);

        u32 config_version;
        if (!ini.Get("Version", "Config", &config_version) || config_version != CONFIG_VERSION)
        {
            // save file and reload
            ini.Clear();

#define SETTING(_type, _var, _section, _defaultval) \
    ini.Set(_section, #_var, _defaultval)
#include "Settings.def"
#undef SETTING

            ini.Set("Version", "Config", CONFIG_VERSION);
            ini.Save(ini_filename);
            ini.Load(ini_filename);
        }

#define SETTING(_type, _var, _section, _defaultval) \
    ini.Get(_section, #_var, &##_var)
#include "Settings.def"
#undef SETTING
    }

#define SETTING(_type, _var, _section, _defaultval) \
	private: _type _var; \
	public: const _type& Get##_var() const { return _var; };
#include "Settings.def"
#undef SETTING
};

class Context : NonCopyable
{
public:
    static Context instance;

    Context();
    ~Context();

    bool UsePEC() { return m_config.GetUsePEC(); };
	bool HardcoreMode() { return m_config.GetHardcoreMode(); };
    s32 OnGPUinit();
    s32 OnGPUshutdown();
    s32 OnGPUOpen(HWND hwndGPU);
    s32 OnGPUclose();
    void OnGPUaddVertex(s16 sx, s16 sy, s64 fx, s64 fy, s64 fz);

    s32 OnGPUtest();

    void LoadPad();

    void* GetEmulatorMemory(u32 offset = 0)
    {
        std::string exename;
        ModuleFileName(&exename);
        if (findStringIC(exename, "ePSXeCutor"))
            return nullptr;
        if (!findStringIC(exename, "ePSXe"))
            return nullptr;

        return (u8*)GetModuleHandle(NULL) + offset;
    }

private:
    Config m_config;
    GPUPatches m_gpupatches;
    PADPlugin m_padplugin;
};

static Context& context = Context::instance;