// ****************************************************************************
// * This file is part of the xBRZ project. It is distributed under           *
// * GNU General Public License: http://www.gnu.org/licenses/gpl-3.0          *
// * Copyright (C) Zenju (zenju AT gmx DOT de) - All Rights Reserved          *
// ****************************************************************************

#include "stdafx.h"
#include <stdlib.h>

#include "../gpuPeteOpenGL2Tweak.h"
#include "hq.h"
#include "xbrz.h"
#include <ppl.h>

static const int TASK_GRANULARITY = 32;
static const xbrz::ScalerCfg xbrzcfg;
void parallelScaleXBRZ(size_t factor, const uint32_t* src, uint32_t* trg, int srcWidth, int srcHeight, xbrz::ColorFormat format)
{
	concurrency::parallel_for(0, srcHeight, TASK_GRANULARITY, [&](const int& i)
	{
		xbrz::scale(factor, src, trg, srcWidth, srcHeight, format, xbrzcfg, i, i + TASK_GRANULARITY);
	});
}

#pragma pack(push, 1)
struct RGB {
	unsigned char r, g, b;
};

struct RGBA {
	unsigned char r, g, b, a;
};

struct ARGB {
	unsigned char a, r, g, b;
};
#pragma pack(pop)

void RGB8ToRGBA8(void* out, const void* in, int size)
{
	const RGB* src = (const RGB*)in;
	RGBA* dst = (RGBA*)out;
	do {
		dst->r = src->r;
		dst->g = src->g;
		dst->b = src->b;
		dst->a = 0xFF;
		src++;
		dst++;
	} while (--size);
}

void RGBA8ToRGB8(void* out, const void* in, int size)
{
	const RGBA* src = (const RGBA*)in;
	RGB* dst = (RGB*)out;
	do {
		dst->r = src->r;
		dst->g = src->g;
		dst->b = src->b;
		src++;
		dst++;
	} while (--size);
}


void scaleXbrz(int scale, const void* srcPtr, unsigned char* dstPtr, int bpp, int width, int height)
{
#if 0
	if (bpp == 24)
	{
		uint32_t* buf_rgba = (uint32_t*)malloc(width * height * 4);
		uint32_t* buf_scaled = (uint32_t*)malloc(width * scale * height * scale * 4);
		RGB8ToRGBA8(buf_rgba, srcPtr, width * height);
		parallelScaleXBRZ(scale, buf_rgba, (uint32_t*)buf_scaled, width, height, xbrz::ColorFormat::RGB);
		//xbrz::scale(scale, buf_rgba, buf_scaled, width, height, xbrz::ColorFormat::RGB, xbrz::ScalerCfg());
		RGBA8ToRGB8(dstPtr, buf_scaled, width * scale * height * scale);
		free(buf_rgba);
		free(buf_scaled);
	}
	else if (bpp == 32)
#endif // 0

		parallelScaleXBRZ(scale, (uint32_t*)srcPtr, (uint32_t*)dstPtr, width, height, xbrz::ColorFormat::ARGB);
		//xbrz::scale(scale, (uint32_t*)srcPtr, (uint32_t*)dstPtr, width, height, xbrz::ColorFormat::ARGB, xbrzcfg);
}

